@@ -31,16 +31,19 @@
 #include <config.h>
 #endif
 
+#include <string.h> /* For strlen */
+
+#include <glib/gutils.h>
 #include <libgnome/gnome-i18n.h>
 
 #include <gedit-menus.h>
 #include <gedit-plugin.h>
 #include <gedit-debug.h>
 
-#define MENU_ITEM_LABEL		N_("_Hello World")
-#define MENU_ITEM_PATH		"/menu/Tools/ToolsOps/"
-#define MENU_ITEM_NAME		"HelloWorld"	
-#define MENU_ITEM_TIP		N_("Prints Hello World.")
+#define MENU_ITEM_LABEL		N_("Insert User Na_me")
+#define MENU_ITEM_PATH		"/menu/Edit/EditOps_4/"
+#define MENU_ITEM_NAME		"UserName"	
+#define MENU_ITEM_TIP		N_("Insert the user name at the cursor position")
 
 G_MODULE_EXPORT GeditPluginState update_ui (GeditPlugin *plugin, BonoboWindow *window);
 G_MODULE_EXPORT GeditPluginState destroy (GeditPlugin *pd);
@@ -50,10 +53,13 @@ G_MODULE_EXPORT GeditPluginState init (GeditPlugin *pd);
 
 
 static void
-hello_world_cb (BonoboUIComponent *uic, gpointer user_data, const gchar* verbname)
+sample_cb (BonoboUIComponent *uic, gpointer user_data, const gchar* verbname)
 {
 	GeditDocument *doc;
 	GeditView *view;
+	gchar *user_name;
+	gchar *user_name_utf8;
+	const gchar *temp;
 	
 	gedit_debug (DEBUG_PLUGINS, "");
 
@@ -63,11 +69,36 @@ hello_world_cb (BonoboUIComponent *uic, gpointer user_data, const gchar* verbnam
 	doc = gedit_view_get_document (view);
 	g_return_if_fail (doc != NULL);
 
+	temp = g_get_real_name ();
+	g_return_if_fail (temp != NULL);
+	
+	if (strlen (temp) <= 0)
+	{
+		temp = g_get_user_name ();
+		g_return_if_fail (temp != NULL);
+	}
+	
+	user_name = g_strdup_printf ("%s ", temp);
+	g_return_if_fail (user_name != NULL);
+
+	if (g_utf8_validate (user_name, -1, NULL))
+		user_name_utf8 = user_name;
+	else
+	{
+		user_name_utf8 = g_locale_to_utf8 (user_name, -1, NULL, NULL, NULL);
+		g_free (user_name);
+
+		if (user_name_utf8 == NULL)
+			user_name_utf8 = g_strdup (" ");
+	}
+
 	gedit_document_begin_user_action (doc);
 	
-	gedit_document_insert_text_at_cursor (doc, _("Hello World "), -1);
+	gedit_document_insert_text_at_cursor (doc, user_name_utf8, -1);
 
 	gedit_document_end_user_action (doc);
+
+	g_free (user_name_utf8);
 }
 
 G_MODULE_EXPORT GeditPluginState
@@ -114,7 +145,7 @@ activate (GeditPlugin *pd)
 		gedit_menus_add_menu_item (BONOBO_WINDOW (top_windows->data),
 				     MENU_ITEM_PATH, MENU_ITEM_NAME,
 				     MENU_ITEM_LABEL, MENU_ITEM_TIP, NULL,
-				     hello_world_cb);
+				     sample_cb);
 
                 pd->update_ui (pd, BONOBO_WINDOW (top_windows->data));
 
@@ -137,12 +168,7 @@ init (GeditPlugin *pd)
 {
 	/* initialize */
 	gedit_debug (DEBUG_PLUGINS, "");
-     
-	pd->name = _("Hello World");
-	pd->desc = _("Sample 'hello world' plugin.");
-	pd->author = "Paolo Maggi <maggi@athena.polito.it>";
-	pd->copyright = _("Copyright (C) 2002 - Paolo Maggi");
-	
+
 	pd->private_data = NULL;
 		
 	return PLUGIN_OK;
