@@ -42,9 +42,9 @@
 
 #include <libgnome/libgnome.h>
 #include <libgnomevfs/gnome-vfs.h>
+#include <eel/eel-vfs-extensions.h>
 
-#include "gedit-prefs.h"
-#include "gnome-vfs-helpers.h"
+#include "gedit-prefs-manager.h"
 #include "gedit-document.h"
 #include "gedit-debug.h"
 #include "gedit-utils.h"
@@ -52,8 +52,6 @@
 
 #include "gedit-marshal.h"
 
-#define NOT_EDITABLE_TAG_NAME "not_editable_tag"
-
 #ifdef MAXPATHLEN
 #define GEDIT_MAX_PATH_LEN  MAXPATHLEN
 #elif defined (PATH_MAX)
@@ -67,9 +65,12 @@ struct _GeditDocumentPrivate
 	gchar		*uri;
 	gint 		 untitled_number;	
 
+	gchar		*encoding;
+
 	gchar		*last_searched_text;
 	gchar		*last_replace_text;
 	gboolean  	 last_search_was_case_sensitive;
+	gboolean  	 last_search_was_entire_word;
 
 	guint		 auto_save_timeout;
 	gboolean	 last_save_was_manually; 
@@ -89,8 +90,6 @@ enum {
 	LAST_SIGNAL
 };
 
-static void gedit_document_base_init 		(GeditDocumentClass 	*klass);
-static void gedit_document_base_finalize	(GeditDocumentClass 	*klass);
 static void gedit_document_class_init 		(GeditDocumentClass 	*klass);
 static void gedit_document_init 		(GeditDocument 		*document);
 static void gedit_document_finalize 		(GObject 		*object);
@@ -173,8 +172,8 @@ gedit_document_get_type (void)
       		static const GTypeInfo our_info =
       		{
         		sizeof (GeditDocumentClass),
-        		(GBaseInitFunc) gedit_document_base_init,
-        		(GBaseFinalizeFunc) gedit_document_base_finalize,
+        		NULL, 		/* base_init,*/
+        		NULL, 		/* base_finalize, */
         		(GClassInitFunc) gedit_document_class_init,
         		NULL,           /* class_finalize */
         		NULL,           /* class_data */
@@ -192,25 +191,6 @@ gedit_document_get_type (void)
 	return document_type;
 }
 
-static void
-gedit_document_base_init (GeditDocumentClass *klass)
-{
-	GtkTextTag *not_editable_tag;
-
-	klass->tag_table = gtk_text_tag_table_new ();
-
-	not_editable_tag =  gtk_text_tag_new (NOT_EDITABLE_TAG_NAME);	
-	g_object_set (G_OBJECT (not_editable_tag), "editable", FALSE, NULL);
-	
-	gtk_text_tag_table_add (klass->tag_table, not_editable_tag);	
-}
-
-static void
-gedit_document_base_finalize (GeditDocumentClass *klass)
-{
-	g_object_unref (G_OBJECT (klass->tag_table));
-	klass->tag_table = NULL;
-}
 	
 static void
 gedit_document_class_init (GeditDocumentClass *klass)
@@ -298,10 +278,6 @@ gedit_document_init (GeditDocument *document)
 {
 	gedit_debug (DEBUG_DOCUMENT, "");
 
-	g_return_if_fail (document->buffer.tag_table == NULL);
-	document->buffer.tag_table = (GEDIT_DOCUMENT_GET_CLASS (document))->tag_table;
-	g_object_ref (G_OBJECT (document->buffer.tag_table));
-		
 	document->priv = g_new0 (GeditDocumentPrivate, 1);
 	
 	document->priv->uri = NULL;
@@ -313,6 +289,17 @@ gedit_document_init (GeditDocument *document)
 
 	document->priv->undo_manager = gedit_undo_manager_new (document);
 
+	if (gedit_prefs_manager_get_save_encoding () == 
+			GEDIT_SAVE_ORIGINAL_FILE_ENCODING_IF_POSSIBLE_NCL)
+	{
+		const gchar *encoding = NULL;
+		g_get_charset(&encoding);
+				
+		document->priv->encoding = g_strdup (encoding);
+	}
+	else
+		document->priv->encoding = NULL;
+			
 	g_signal_connect (G_OBJECT (document->priv->undo_manager), "can_undo",
 			  G_CALLBACK (gedit_document_can_undo_handler), 
 			  document);
@@ -346,18 +333,10 @@ gedit_document_finalize (GObject *object)
 				document->priv->untitled_number);
 	}
 
-	if (document->priv->uri)
-    	{
-		g_free (document->priv->uri);
-      		document->priv->uri = NULL;
-	}		
-	
-	if (document->priv->last_searched_text)
-		g_free (document->priv->last_searched_text);
-
-	if (document->priv->last_replace_text)
-		g_free (document->priv->last_replace_text);
-
+	g_free (document->priv->uri);
+	g_free (document->priv->last_searched_text);
+	g_free (document->priv->last_replace_text);
+	g_free (document->priv->encoding);
 
 	g_object_unref (G_OBJECT (document->priv->undo_manager));
 	
@@ -434,14 +413,18 @@ gedit_document_new_with_uri (const gchar *uri, GError **error)
 void 		
 gedit_document_set_readonly (GeditDocument *document, gboolean readonly)
 {
+	gboolean auto_save;
+	
 	gedit_debug (DEBUG_DOCUMENT, "");
 
 	g_return_if_fail (document != NULL);
 	g_return_if_fail (document->priv != NULL);
 
+	auto_save = gedit_prefs_manager_get_auto_save ();
+	
 	if (readonly) 
 	{
-		if (gedit_settings->auto_save && (document->priv->auto_save_timeout > 0))
+		if (auto_save && (document->priv->auto_save_timeout > 0))
 		{
 			gedit_debug (DEBUG_DOCUMENT, "Remove autosave timeout");
 
@@ -451,12 +434,17 @@ gedit_document_set_readonly (GeditDocument *document, gboolean readonly)
 	}
 	else
 	{
-		if (gedit_settings->auto_save && (document->priv->auto_save_timeout <= 0))
+		if (auto_save && (document->priv->auto_save_timeout <= 0))
 		{
+			gint auto_save_interval;
+			
 			gedit_debug (DEBUG_DOCUMENT, "Install autosave timeout");
 
+			auto_save_interval = 
+				gedit_prefs_manager_get_auto_save_interval ();
+				
 			document->priv->auto_save_timeout = g_timeout_add 
-				(gedit_settings->auto_save_interval*1000*60,
+				(auto_save_interval * 1000 * 60,
 		 		 (GSourceFunc)gedit_document_auto_save_timeout,
 		  		 document);		
 		}
@@ -550,15 +538,10 @@ gedit_document_get_uri (const GeditDocument* doc)
 	else
 	{
 		gchar *res;
-		gchar *uri = g_filename_to_utf8 (doc->priv->uri, -1, NULL, NULL, NULL);
 
-		if (uri == NULL)
-			return g_strdup (_("Invalid file name"));
+		res = eel_format_uri_for_display (doc->priv->uri);
+		g_return_val_if_fail (res != NULL, g_strdup (_("Invalid file name")));
 		
-		res = gnome_vfs_x_format_uri_for_display (uri);
-
-		g_free (uri);
-
 		return res;
 	}
 }
@@ -572,17 +555,39 @@ gedit_document_get_short_name (const GeditDocument* doc)
 		return g_strdup_printf (_("%s %d"), _("Untitled"), doc->priv->untitled_number);
 	else
 	{
-		gchar *res;
-		gchar *uri = g_filename_to_utf8 (doc->priv->uri, -1, NULL, NULL, NULL);
+		gchar *basename;
+		gchar *utf8_basename;
 		
-		if (uri == NULL)
-			return g_strdup (_("Invalid file name"));
+		basename = eel_uri_get_basename (doc->priv->uri);
 
-		res = gnome_vfs_x_uri_get_basename (uri);
+		if (basename != NULL) 
+		{
+			gboolean filenames_are_locale_encoded;
+		       	filenames_are_locale_encoded = g_getenv ("G_BROKEN_FILENAMES") != NULL;
 
-		g_free (uri);
-
-		return res;
+			if (filenames_are_locale_encoded) 
+			{
+				utf8_basename = g_locale_to_utf8 (basename, -1, NULL, NULL, NULL);
+				
+				if (utf8_basename != NULL) 
+				{
+					g_free (basename);
+					return utf8_basename;
+				} 
+			}
+			else 
+			{
+				if (g_utf8_validate (basename, -1, NULL)) 
+					return basename;
+			}
+			
+			/* there are problems */
+			utf8_basename = eel_make_valid_utf8 (basename);
+			g_free (basename);
+			return utf8_basename;
+		}
+		else
+			return 	g_strdup (_("Invalid file name"));
 	}
 }
 
@@ -606,8 +611,8 @@ gedit_document_auto_save_timeout (GeditDocument *doc)
 
 	g_return_val_if_fail (!gedit_document_is_readonly (doc), FALSE);
 
-	/* Romove timeout if now gedit_settings->auto_save is FALSE */
-	if (!gedit_settings->auto_save) 
+	/* Remove timeout if now auto_save is FALSE */
+	if (!gedit_prefs_manager_get_auto_save ()) 
 		return FALSE;
 
 	if (!gedit_document_get_modified (doc))
@@ -662,7 +667,7 @@ gedit_document_load (GeditDocument* doc, const gchar *uri, GError **error)
 	g_return_val_if_fail (doc != NULL, FALSE);
 	g_return_val_if_fail (uri != NULL, FALSE);
 
-	res = gnome_vfs_x_read_entire_file (uri, &file_size, &file_contents);
+	res = eel_read_entire_file (uri, &file_size, &file_contents);
 
 	gedit_debug (DEBUG_DOCUMENT, "End reading %s (result: %s)", uri, gnome_vfs_result_to_string (res));
 	
@@ -675,44 +680,40 @@ gedit_document_load (GeditDocument* doc, const gchar *uri, GError **error)
 
 	if (file_size > 0)
 	{
-		if (!g_utf8_validate (file_contents, file_size, NULL))
+		gchar *converted_text;
+		gint len = -1;
+
+		if (g_utf8_validate (file_contents, file_size, NULL))
 		{
-			/* The file contains invalid UTF8 data */
-			/* Try to convert it to UTF-8 from currence locale */
-			GError *conv_error = NULL;
-			gchar* converted_file_contents = NULL;
-			gsize bytes_written;
-			
-			converted_file_contents = g_locale_to_utf8 (file_contents, file_size,
-					NULL, &bytes_written, &conv_error); 
-			
+			converted_text = file_contents;
+			len = file_size;
+			file_contents = NULL;
+		}
+		else
+			converted_text = gedit_utils_convert_to_utf8 (file_contents,
+							file_size,
+							&doc->priv->encoding);
+
+		if (converted_text == NULL)
+		{
+			/* bail out */
+			g_set_error (error, GEDIT_DOCUMENT_IO_ERROR,
+				     GEDIT_ERROR_INVALID_UTF8_DATA,
+				     _("Invalid UTF-8 data"));
+
 			g_free (file_contents);
 
-			if ((conv_error != NULL) || 
-			    !g_utf8_validate (converted_file_contents, bytes_written, NULL))		
-			{
-				/* Coversion failed */	
-				if (conv_error != NULL)
-					g_error_free (conv_error);
-
-				g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, 
-					     GEDIT_ERROR_INVALID_UTF8_DATA,
-					     _("Invalid UTF-8 data"));
-				
-				if (converted_file_contents != NULL)
-					g_free (converted_file_contents);
-				
-				return FALSE;
-			}
-
-			file_contents = converted_file_contents;
-			file_size = bytes_written;
+			return FALSE;
 		}
 
+		gedit_debug (DEBUG_DOCUMENT, "Document encoding: %s", 
+			     doc->priv->encoding == NULL ? "UTF-8 (Null)" : doc->priv->encoding);
+
 		gedit_undo_manager_begin_not_undoable_action (doc->priv->undo_manager);
 		/* Insert text in the buffer */
 		gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &iter, 0);
-		gtk_text_buffer_insert (GTK_TEXT_BUFFER (doc), &iter, file_contents, file_size);
+		gtk_text_buffer_insert (GTK_TEXT_BUFFER (doc), &iter, converted_text, len);
+		g_free (converted_text);
 
 		/* We had a newline in the buffer to begin with. (The buffer always contains
    		 * a newline, so we delete to the end of the buffer to clean up. */
@@ -726,7 +727,8 @@ gedit_document_load (GeditDocument* doc, const gchar *uri, GError **error)
 		gedit_undo_manager_end_not_undoable_action (doc->priv->undo_manager);
 	}
 
-	g_free (file_contents);
+	if (file_contents != NULL)
+		g_free (file_contents);
 
 	if (gedit_utils_is_uri_read_only (uri))
 	{
@@ -750,68 +752,48 @@ gedit_document_load (GeditDocument* doc, const gchar *uri, GError **error)
 	return TRUE;
 }
 
-#define GEDIT_STDIN_BUFSIZE 1024
-
 gboolean
 gedit_document_load_from_stdin (GeditDocument* doc, GError **error)
 {
-	GString * file_contents;
-	gchar *tmp_buf = NULL;
-	struct stat stats;
-	guint buffer_length;
-
+	gchar *stdin_data;
 	GtkTextIter iter, end;
-	GnomeVFSResult	res;
 	
 	gedit_debug (DEBUG_DOCUMENT, "");
 
 	g_return_val_if_fail (doc != NULL, FALSE);
 	
-	fstat (STDIN_FILENO, &stats);
-	
-	if (stats.st_size  == 0)
-		return FALSE;
+	stdin_data = gedit_utils_get_stdin ();
 
-	tmp_buf = g_new0 (gchar, GEDIT_STDIN_BUFSIZE + 1);
-	g_return_val_if_fail (tmp_buf != NULL, FALSE);
-
-	file_contents = g_string_new (NULL);
-	
-	while (feof (stdin) == 0)
+	if (stdin_data != NULL && strlen (stdin_data) > 0)
 	{
-		buffer_length = fread (tmp_buf, 1, GEDIT_STDIN_BUFSIZE, stdin);
-		tmp_buf [buffer_length + 1] = '\0';
-		g_string_append (file_contents, tmp_buf);
+		gchar *converted_text;
 
-		if (ferror (stdin) != 0)
+		if (g_utf8_validate (stdin_data, -1, NULL))
 		{
-			res = gnome_vfs_result_from_errno (); 
-		
-			g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, res,
-				gnome_vfs_result_to_string (res));
-
-			g_free (tmp_buf);
-			g_string_free (file_contents, TRUE);
-			return FALSE;
+			converted_text = stdin_data;
+			stdin_data = NULL;
 		}
-	}
+		else
+			converted_text = gedit_utils_convert_to_utf8 (stdin_data,
+							-1,
+							&doc->priv->encoding);
 
-	fclose (stdin);
-
-	if (file_contents->len > 0)
-	{
-		if (!g_utf8_validate (file_contents->str, file_contents->len, NULL))
+		if (converted_text == NULL)
 		{
-			g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, GEDIT_ERROR_INVALID_UTF8_DATA,
-				_("Invalid UTF-8 data"));
-			g_string_free (file_contents, TRUE);
+			/* bail out */
+			g_set_error (error, GEDIT_DOCUMENT_IO_ERROR,
+				     GEDIT_ERROR_INVALID_UTF8_DATA,
+				     _("Invalid UTF-8 data"));
+
+			g_free (stdin_data);
+
 			return FALSE;
 		}
 
 		gedit_undo_manager_begin_not_undoable_action (doc->priv->undo_manager);
 		/* Insert text in the buffer */
 		gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &iter, 0);
-		gtk_text_buffer_insert (GTK_TEXT_BUFFER (doc), &iter, file_contents->str, file_contents->len);
+		gtk_text_buffer_insert (GTK_TEXT_BUFFER (doc), &iter, converted_text, -1);
 
 		/* We had a newline in the buffer to begin with. (The buffer always contains
    		 * a newline, so we delete to the end of the buffer to clean up. */
@@ -823,9 +805,13 @@ gedit_document_load_from_stdin (GeditDocument* doc, GError **error)
 		gtk_text_buffer_place_cursor (GTK_TEXT_BUFFER (doc), &iter);
 
 		gedit_undo_manager_end_not_undoable_action (doc->priv->undo_manager);
+		
+		g_free (converted_text);
 	}
 
-	g_string_free (file_contents, TRUE);
+	if (stdin_data != NULL)
+		g_free (stdin_data);
+			
 
 	gedit_document_set_readonly (doc, FALSE);
 	gtk_text_buffer_set_modified (GTK_TEXT_BUFFER (doc), TRUE);
@@ -873,14 +859,22 @@ gedit_document_set_uri (GeditDocument* doc, const gchar* uri)
 gboolean 
 gedit_document_save (GeditDocument* doc, GError **error)
 {	
+	gboolean auto_save;
+	gboolean create_backup_copy;
+
 	gboolean ret;
+
 	gedit_debug (DEBUG_DOCUMENT, "");
 
 	g_return_val_if_fail (doc != NULL, FALSE);
 	g_return_val_if_fail (doc->priv != NULL, FALSE);
 	g_return_val_if_fail (!doc->priv->readonly, FALSE);
-
-	if (gedit_settings->auto_save) 
+	g_return_val_if_fail (doc->priv->uri != NULL, FALSE);
+	
+	auto_save = gedit_prefs_manager_get_auto_save ();
+	create_backup_copy = gedit_prefs_manager_get_create_backup_copy ();
+		
+	if (auto_save) 
 	{
 		if (doc->priv->auto_save_timeout > 0)
 		{
@@ -889,17 +883,23 @@ gedit_document_save (GeditDocument* doc, GError **error)
 		}
 	}
 	
-	ret = gedit_document_save_as_real (doc, doc->priv->uri, 
-			gedit_settings->create_backup_copy, error);
+	ret = gedit_document_save_as_real (doc, 
+					   doc->priv->uri, 
+					   create_backup_copy, 
+					   error);
 
 	if (ret)
 		doc->priv->last_save_was_manually = TRUE;
 
-	if (gedit_settings->auto_save && (doc->priv->auto_save_timeout <= 0)) 
+	if (auto_save && (doc->priv->auto_save_timeout <= 0)) 
 	{
+		gint auto_save_interval = 
+			gedit_prefs_manager_get_auto_save_interval ();
+
 		doc->priv->auto_save_timeout =
-			g_timeout_add (gedit_settings->auto_save_interval*1000*60,
-				       (GSourceFunc)gedit_document_auto_save, doc);
+			g_timeout_add (auto_save_interval * 1000 * 60,
+				       (GSourceFunc) gedit_document_auto_save, 
+				       doc);
 	}
 
 	return ret;
@@ -908,6 +908,8 @@ gedit_document_save (GeditDocument* doc, GError **error)
 gboolean	
 gedit_document_save_as (GeditDocument* doc, const gchar *uri, GError **error)
 {	
+	gboolean auto_save;
+	
 	gboolean ret = FALSE;
 	
 	gedit_debug (DEBUG_DOCUMENT, "");
@@ -916,7 +918,9 @@ gedit_document_save_as (GeditDocument* doc, const gchar *uri, GError **error)
 	g_return_val_if_fail (doc->priv != NULL, FALSE);
 	g_return_val_if_fail (uri != NULL, FALSE);
 
-	if (gedit_settings->auto_save) 
+	auto_save = gedit_prefs_manager_get_auto_save ();
+
+	if (auto_save) 
 	{
 
 		if (doc->priv->auto_save_timeout > 0)
@@ -926,7 +930,7 @@ gedit_document_save_as (GeditDocument* doc, const gchar *uri, GError **error)
 		}
 	}
 
-	if (gedit_document_save_as_real (doc, uri, FALSE, error))
+	if (gedit_document_save_as_real (doc, uri, TRUE, error))
 	{
 		gedit_document_set_uri (doc, uri);
 		gedit_document_set_readonly (doc, FALSE);
@@ -935,10 +939,13 @@ gedit_document_save_as (GeditDocument* doc, const gchar *uri, GError **error)
 		ret = TRUE;
 	}		
 
-	if (gedit_settings->auto_save && (doc->priv->auto_save_timeout <= 0)) 
+	if (auto_save && (doc->priv->auto_save_timeout <= 0)) 
 	{
+		gint auto_save_interval =
+			gedit_prefs_manager_get_auto_save_interval ();
+		
 		doc->priv->auto_save_timeout =
-			g_timeout_add (gedit_settings->auto_save_interval*1000*60,
+			g_timeout_add (auto_save_interval * 1000 * 60,
 				       (GSourceFunc)gedit_document_auto_save, doc);
 	}
 
@@ -975,6 +982,9 @@ follow_symlinks (const gchar *filename, GError **error)
 	gchar *followed_filename;
 	gint link_count;
 
+	g_return_val_if_fail (filename != NULL, NULL);
+	
+	gedit_debug (DEBUG_DOCUMENT, "Filename: %s", filename); 
 	g_return_val_if_fail (strlen (filename) + 1 <= GEDIT_MAX_PATH_LEN, NULL);
 
 	followed_filename = g_strdup (filename);
@@ -1070,11 +1080,14 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 	gchar *dirname;
 	mode_t saved_umask;
 	struct stat st;
-	char *chars;
+	gchar *chars = NULL;
 	gint chars_len;
 	gint fd;
 	gint retval;
-
+	GeditSaveEncodingSetting encoding_setting;
+	gboolean res;
+	gboolean add_cr;
+	
 	gedit_debug (DEBUG_DOCUMENT, "");
 
 	g_return_val_if_fail (doc != NULL, FALSE);
@@ -1093,8 +1106,12 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 	if (!gedit_utils_uri_has_file_scheme (uri))
 	{
 		gchar *error_message;
-		gchar *scheme_string = gnome_vfs_x_uri_get_scheme (uri);
-                
+		gchar *scheme_string;
+
+		gchar *temp = eel_uri_get_scheme (uri);
+                scheme_string = eel_make_valid_utf8 (temp);
+		g_free (temp);
+
 		if (scheme_string != NULL)
 		{
 			error_message = g_strdup_printf (
@@ -1112,7 +1129,8 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 		return FALSE;
 	}
 
-	filename = gnome_vfs_x_format_uri_for_display (uri);
+	/* Get filename from uri */
+	filename = gnome_vfs_get_local_path_from_uri (uri);
 
 	if (!filename)
 	{
@@ -1128,20 +1146,7 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 	if (!real_filename)
 		goto out;
 
-	if (stat (real_filename, &st) != 0)
-	{
-		/* File does not exist? */
-		create_backup_copy = FALSE;
-
-		/* Use default permissions */
-		st.st_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
-		st.st_uid = getuid ();
-		st.st_gid = getgid ();
-	}
-
-	/* Save to a temporary file.  We set the umask because some (buggy)
-	 * implementations of mkstemp() use permissions 0666 and we want 0600.
-	 */
+	/* Get the directory in which the real filename lives */
 
 	slashpos = strrchr (real_filename, G_DIR_SEPARATOR);
 
@@ -1153,6 +1158,36 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 	else
 		dirname = g_strdup (".");
 
+	/* If there is not an existing file with that name, compute the
+	 * permissions and uid/gid that we will use for the newly-created file.
+	 */
+
+	if (stat (real_filename, &st) != 0)
+	{
+		struct stat dir_st;
+		int result;
+
+		/* File does not exist? */
+		create_backup_copy = FALSE;
+
+		/* Use default permissions */
+		saved_umask = umask(0);
+		st.st_mode = 0666 & ~saved_umask;
+		umask(saved_umask);
+		st.st_uid = getuid ();
+
+		result = stat (dirname, &dir_st);
+
+		if (result == 0 && (dir_st.st_mode & S_ISGID))
+			st.st_gid = dir_st.st_gid;
+		else
+			st.st_gid = getgid ();
+	}
+
+	/* Save to a temporary file.  We set the umask because some (buggy)
+	 * implementations of mkstemp() use permissions 0666 and we want 0600.
+	 */
+
 	temp_filename = g_build_filename (dirname, ".gedit-save-XXXXXX", NULL);
 	g_free (dirname);
 
@@ -1166,13 +1201,17 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 		goto out;
 	}
 
-      	chars = gedit_document_get_buffer (doc);
+	chars = gedit_document_get_chars (doc, 0, -1);
 
-	if (gedit_settings->save_encoding == GEDIT_SAVE_CURRENT_LOCALE_WHEN_POSSIBLE)
+	encoding_setting = gedit_prefs_manager_get_save_encoding ();
+
+	if (encoding_setting == GEDIT_SAVE_CURRENT_LOCALE_IF_POSSIBLE)
 	{
 		GError *conv_error = NULL;
 		gchar* converted_file_contents = NULL;
 
+		gedit_debug (DEBUG_DOCUMENT, "Using current locale's encoding");
+
 		converted_file_contents = g_locale_from_utf8 (chars, -1, NULL, NULL, &conv_error);
 
 		if (conv_error != NULL)
@@ -1186,9 +1225,50 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 			chars = converted_file_contents;
 		}
 	}
+	else
+	{
+		if ((doc->priv->encoding != NULL) &&
+		    ((encoding_setting == GEDIT_SAVE_ORIGINAL_FILE_ENCODING_IF_POSSIBLE) ||
+		     (encoding_setting == GEDIT_SAVE_ORIGINAL_FILE_ENCODING_IF_POSSIBLE_NCL)))
+		{
+			GError *conv_error = NULL;
+			gchar* converted_file_contents = NULL;
+
+			gedit_debug (DEBUG_DOCUMENT, "Using encoding %s", doc->priv->encoding);
+
+			/* Try to convert it from UTF-8 to original encoding */
+			converted_file_contents = g_convert (chars, -1, 
+							     doc->priv->encoding, "UTF-8", 
+							     NULL, NULL, &conv_error); 
+
+			if (conv_error != NULL)
+			{
+				/* Conversion error */
+				g_error_free (conv_error);
+			}
+			else
+			{
+				g_free (chars);
+				chars = converted_file_contents;
+			}
+		}
+		else
+			gedit_debug (DEBUG_DOCUMENT, "Using UTF-8 (Null)");
+
+	}	    
 
 	chars_len = strlen (chars);
-	if (write (fd, chars, chars_len) != chars_len)
+
+	add_cr = (*(chars + chars_len - 1) != '\n');
+
+	/* Save the file content */
+	res = (write (fd, chars, chars_len) == chars_len);
+
+	if (res && add_cr)
+		/* Add \n if needed */
+		res = (write (fd, "\n", 1) == 1);
+
+	if (!res)
 	{
 		gchar *msg;
 
@@ -1228,10 +1308,17 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 
 	if (create_backup_copy)
 	{
+		gchar *bak_ext;
 		gint result;
 
-		backup_filename = g_strconcat (real_filename, gedit_settings->backup_extension, NULL);
+		bak_ext = gedit_prefs_manager_get_backup_extension ();
+		
+		backup_filename = g_strconcat (real_filename, 
+					       bak_ext,
+					       NULL);
 
+		g_free (bak_ext);
+		
 		result = rename (real_filename, backup_filename);
 
 		if (result != 0)
@@ -1252,7 +1339,7 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 		saved_errno = errno;
 
 		if (create_backup_copy && rename (backup_filename, real_filename) != 0)
-			g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, saved_errno,
+			g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, errno,
 				     " ");
 		else
 			g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, saved_errno,
@@ -1276,6 +1363,8 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 
  out:
 
+	g_free (chars);
+
 	g_free (filename);
 	g_free (real_filename);
 	g_free (backup_filename);
@@ -1284,21 +1373,6 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 	return retval;
 }
 
-gchar* 
-gedit_document_get_buffer (const GeditDocument *doc)
-{
-	GtkTextIter start, end;
-
-	gedit_debug (DEBUG_DOCUMENT, "");
-
-	g_return_val_if_fail (doc != NULL, FALSE);
-
-	gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &start, 0);
-      	gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (doc), &end);
-  
-      	return gtk_text_buffer_get_slice (GTK_TEXT_BUFFER (doc), &start, &end, TRUE);
-}
-
 gboolean	
 gedit_document_is_untitled (const GeditDocument* doc)
 {
@@ -1358,21 +1432,6 @@ gedit_document_get_line_count (const GeditDocument *doc)
 	return gtk_text_buffer_get_line_count (GTK_TEXT_BUFFER (doc));
 }
 
-void 
-gedit_document_delete_all_text (GeditDocument *doc)
-{
-	GtkTextIter start, end;
-
-	gedit_debug (DEBUG_DOCUMENT, "");
-
-	g_return_if_fail (doc != NULL);
-
-	gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &start, 0);
-      	gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (doc), &end);
-	
-	gtk_text_buffer_delete (GTK_TEXT_BUFFER (doc), &start, &end);
-}
-
 gboolean 
 gedit_document_revert (GeditDocument *doc,  GError **error)
 {
@@ -1389,11 +1448,11 @@ gedit_document_revert (GeditDocument *doc,  GError **error)
 		return FALSE;
 	}
 			
-	buffer = gedit_document_get_buffer (doc);
+	buffer = gedit_document_get_chars (doc, 0, -1);
 
 	gedit_undo_manager_begin_not_undoable_action (doc->priv->undo_manager);
 
-	gedit_document_delete_all_text (doc);
+	gedit_document_delete_text (doc, 0, -1);
 
 	if (!gedit_document_load (doc, doc->priv->uri, error))
 	{
@@ -1457,10 +1516,14 @@ gedit_document_delete_text (GeditDocument *doc, gint start, gint end)
 
 	g_return_if_fail (GEDIT_IS_DOCUMENT (doc));
 	g_return_if_fail (start >= 0);
-	g_return_if_fail (end >= 0);
+	g_return_if_fail ((end > start) || end < 0);
 
 	gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &start_iter, start);
-	gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &end_iter, end);
+
+	if (end < 0)
+		gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (doc), &end_iter);
+	else
+		gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &end_iter, end);
 
 	gtk_text_buffer_delete (GTK_TEXT_BUFFER (doc), &start_iter, &end_iter);
 }
@@ -1475,10 +1538,14 @@ gedit_document_get_chars (GeditDocument *doc, gint start, gint end)
 
 	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), NULL);
 	g_return_val_if_fail (start >= 0, NULL);
-	g_return_val_if_fail (end >= 0, NULL);
+	g_return_val_if_fail ((end > start) || (end < 0), NULL);
 
 	gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &start_iter, start);
-	gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &end_iter, end);
+
+	if (end < 0)
+		gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (doc), &end_iter);
+	else
+		gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &end_iter, end);
 
 	return gtk_text_buffer_get_slice (GTK_TEXT_BUFFER (doc), &start_iter, &end_iter, TRUE);
 }
@@ -1641,13 +1708,44 @@ gedit_document_get_last_replace_text (GeditDocument* doc)
 		g_strdup (doc->priv->last_replace_text) : NULL;	
 }
 
+void 
+gedit_document_set_last_searched_text (GeditDocument* doc, const gchar *text)
+{
+	gedit_debug (DEBUG_DOCUMENT, "");
+
+	g_return_if_fail (GEDIT_IS_DOCUMENT (doc));
+	g_return_if_fail (doc->priv != NULL);
+	g_return_if_fail (text != NULL);
+
+	if (doc->priv->last_searched_text != NULL)
+		g_free (doc->priv->last_searched_text);
+
+	doc->priv->last_searched_text = g_strdup (text);
+}
+
+void 
+gedit_document_set_last_replace_text (GeditDocument* doc, const gchar *text)
+{
+	gedit_debug (DEBUG_DOCUMENT, "");
+
+	g_return_if_fail (GEDIT_IS_DOCUMENT (doc));
+	g_return_if_fail (doc->priv != NULL);
+	g_return_if_fail (text != NULL);
+
+	if (doc->priv->last_replace_text != NULL)
+		g_free (doc->priv->last_replace_text);
+
+	doc->priv->last_replace_text = g_strdup (text);
+}
+
 gboolean
 gedit_document_find (GeditDocument* doc, const gchar* str, 
-		gboolean from_cursor, gboolean case_sensitive)
+		gboolean from_cursor, gboolean case_sensitive, gboolean entire_word)
 {
 	GtkTextIter iter;
 	gboolean found = FALSE;
 	GtkTextSearchFlags search_flags;
+	gchar *converted_str;
 
 	gedit_debug (DEBUG_DOCUMENT, "");
 
@@ -1655,6 +1753,12 @@ gedit_document_find (GeditDocument* doc, const gchar* str,
 	g_return_val_if_fail (doc->priv != NULL, FALSE);
 	g_return_val_if_fail (str != NULL, FALSE);
 
+	converted_str = gedit_utils_convert_search_text (str);
+	g_return_val_if_fail (converted_str != NULL, FALSE);
+	
+	gedit_debug (DEBUG_DOCUMENT, "str: %s", str);
+	gedit_debug (DEBUG_DOCUMENT, "converted_str: %s", converted_str);
+
 	search_flags = GTK_TEXT_SEARCH_VISIBLE_ONLY | GTK_TEXT_SEARCH_TEXT_ONLY;
 	
 	if (!case_sensitive)
@@ -1681,13 +1785,31 @@ gedit_document_find (GeditDocument* doc, const gchar* str,
 	else		
 		gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &iter, 0);
 
-	if (*str != '\0')
+	found = FALSE;
+	
+	if (*converted_str != '\0')
     	{
       		GtkTextIter match_start, match_end;
 
-          	found = gedit_text_iter_forward_search (&iter, str, search_flags,
-                                        	&match_start, &match_end,
-                                               	NULL);	
+		while (!found)
+		{
+          		found = gedit_text_iter_forward_search (&iter, converted_str, search_flags,
+                        	                	&match_start, &match_end,
+                                	               	NULL);	
+
+			if (found && entire_word)
+			{
+				found = gtk_text_iter_starts_word (&match_start) && 
+					gtk_text_iter_ends_word (&match_end);
+
+				if (!found) 
+					iter = match_end;
+
+			}
+			else
+				break;
+		}
+		
 		if (found)
 		{
 			gtk_text_buffer_place_cursor (GTK_TEXT_BUFFER (doc),
@@ -1702,13 +1824,16 @@ gedit_document_find (GeditDocument* doc, const gchar* str,
 
 		doc->priv->last_searched_text = g_strdup (str);
 		doc->priv->last_search_was_case_sensitive = case_sensitive;
+		doc->priv->last_search_was_entire_word = entire_word;
 	}
 
+	g_free (converted_str);
+
 	return found;
 }
 
 gboolean
-gedit_document_find_again (GeditDocument* doc)
+gedit_document_find_again (GeditDocument* doc, gboolean from_cursor)
 {
 	gchar* last_searched_text;
 	gboolean found;
@@ -1723,22 +1848,26 @@ gedit_document_find_again (GeditDocument* doc)
 	if (last_searched_text == NULL)
 		return FALSE;
 	
-	found = gedit_document_find (doc, last_searched_text, TRUE, 
-				doc->priv->last_search_was_case_sensitive);
+	found = gedit_document_find (doc, 
+				     last_searched_text, 
+				     from_cursor, 
+				     doc->priv->last_search_was_case_sensitive, 
+				     doc->priv->last_search_was_entire_word);
+
 	g_free (last_searched_text);
 
 	return found;
 }
 
-gchar*
-gedit_document_get_selected_text (GeditDocument *doc, gint *start, gint *end)
+gboolean 
+gedit_document_get_selection (GeditDocument *doc, gint *start, gint *end)
 {
 	GtkTextIter iter;
 	GtkTextIter sel_bound;
 
 	gedit_debug (DEBUG_DOCUMENT, "");
 
-	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), NULL);
+	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), FALSE);
 
 	gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (doc),			
                                     &iter,
@@ -1757,35 +1886,6 @@ gedit_document_get_selected_text (GeditDocument *doc, gint *start, gint *end)
 	if (end != NULL)
 		*end = gtk_text_iter_get_offset (&sel_bound); 
 
-	if (gtk_text_iter_equal (&sel_bound, &iter))
-	{
-		gedit_debug (DEBUG_DOCUMENT, "There is no selected text");
-
-		return NULL;
-	}
-	return gtk_text_buffer_get_slice (GTK_TEXT_BUFFER (doc), &iter, &sel_bound, TRUE);
-}
-
-gboolean
-gedit_document_has_selected_text (GeditDocument *doc)
-{
-	GtkTextIter iter;
-	GtkTextIter sel_bound;
-
-	gedit_debug (DEBUG_DOCUMENT, "");
-
-	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), FALSE);
-
-	gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (doc),			
-                                    &iter,
-                                    gtk_text_buffer_get_mark (GTK_TEXT_BUFFER (doc),
-					                      "insert"));
-		
-	gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (doc),			
-                                    &sel_bound,
-                                    gtk_text_buffer_get_mark (GTK_TEXT_BUFFER (doc),
-					                      "selection_bound"));
-
 	return !gtk_text_iter_equal (&sel_bound, &iter);	
 }
 
@@ -1794,12 +1894,15 @@ gedit_document_replace_selected_text (GeditDocument *doc, const gchar *replace)
 {
 	GtkTextIter iter;
 	GtkTextIter sel_bound;
+	gchar *converted_replace;
 
 	gedit_debug (DEBUG_DOCUMENT, "");
 
 	g_return_if_fail (GEDIT_IS_DOCUMENT (doc));
 	g_return_if_fail (replace != NULL);	
 
+	converted_replace = gedit_utils_convert_search_text (replace);
+
 	gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (doc),			
                                     &iter,
                                     gtk_text_buffer_get_mark (GTK_TEXT_BUFFER (doc),
@@ -1829,10 +1932,10 @@ gedit_document_replace_selected_text (GeditDocument *doc, const gchar *replace)
                                     &iter,
                                     gtk_text_buffer_get_mark (GTK_TEXT_BUFFER (doc),
 					                      "insert"));
-	if (*replace != '\0')
+	if (*converted_replace != '\0')
 		gtk_text_buffer_insert (GTK_TEXT_BUFFER (doc),
 					&iter,
-					replace, strlen (replace));
+					converted_replace, strlen (converted_replace));
 
 	if (doc->priv->last_replace_text != NULL)
 		g_free (doc->priv->last_replace_text);
@@ -1840,11 +1943,12 @@ gedit_document_replace_selected_text (GeditDocument *doc, const gchar *replace)
 	doc->priv->last_replace_text = g_strdup (replace);
 
 	gedit_document_end_user_action (doc);
+	g_free (converted_replace);
 }
 
 gboolean
 gedit_document_replace_all (GeditDocument *doc,
-	      	const gchar *find, const gchar *replace, gboolean case_sensitive)
+	      	const gchar *find, const gchar *replace, gboolean case_sensitive, gboolean entire_word)
 {
 	gboolean from_cursor = FALSE;
 	gboolean cont = 0;
@@ -1857,7 +1961,7 @@ gedit_document_replace_all (GeditDocument *doc,
 
 	gedit_document_begin_user_action (doc);
 
-	while (gedit_document_find (doc, find, from_cursor, case_sensitive)) 
+	while (gedit_document_find (doc, find, from_cursor, case_sensitive, entire_word)) 
 	{
 		gedit_document_replace_selected_text (doc, replace);
 		
@@ -1913,3 +2017,28 @@ gedit_document_set_cursor (GeditDocument *doc, gint cursor)
 	gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &iter, cursor);
 	gtk_text_buffer_place_cursor (GTK_TEXT_BUFFER (doc), &iter);
 }
+
+void
+gedit_document_set_selection (GeditDocument *doc, gint start, gint end)
+{
+ 	GtkTextIter start_iter;
+	GtkTextIter end_iter;
+
+	gedit_debug (DEBUG_DOCUMENT, "");
+
+	g_return_if_fail (GEDIT_IS_DOCUMENT (doc));
+	g_return_if_fail (start >= 0);
+	g_return_if_fail ((end > start) || (end < 0));
+
+	gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &start_iter, start);
+
+	if (end < 0)
+		gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (doc), &end_iter);
+	else
+		gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (doc), &end_iter, end);
+
+	gtk_text_buffer_place_cursor (GTK_TEXT_BUFFER (doc), &end_iter);
+
+	gtk_text_buffer_move_mark_by_name (GTK_TEXT_BUFFER (doc),
+					"selection_bound", &start_iter);
+}
